from bestia.output import Row, FString, echo, tty_cols

from .keys import *

_NOTE_WIDTH = 5
_FRET_WIDTH = 1

def max_frets_on_screen(limit=24):
    frets = int(
        tty_cols() / ( _NOTE_WIDTH + _FRET_WIDTH )
    ) - 1
    return frets if frets <= limit else limit

class PluckedString(object):

    def __init__(self, tone, alt='', oct=0, frets=0, display=None, verbose=1):

        self.__key = None

        self.tuning = Note(tone, alt, oct)
        self.display_method = display
        self.frets = 12 if not frets else frets
        self.verbose = verbose
 

    @property
    def key(self):
        ''' inits an object of parent Key class from self.display_method '''
        if not self.__key:
            __key  = self.display_method.__self__
            Key = __key.__class__
            self.__key = Key(
                __key.root.chr,
                __key.root.alt,
            )
        return self.__key

    # USERS DO NOT SET STRING.KEY... USE STRING.DISPLAY_METHOD
    # @key.setter
    # def key(self, k):
    #     self.__key = k

    @property
    def display_method(self):
        if not self.__display:
            self.__display = ChromaticKey(*self.tuning).scale
        return self.__display

    @display_method.setter
    def display_method(self, dm):
        self.__display = dm


    @property
    def frets(self):
        return self.__frets

    @frets.setter
    def frets(self, f):
        self.__frets = int(f) +1


    def __repr__(self):
        ''' prints string notes matching given key '''
        string_line = Row()

        for fret_n, note in enumerate(
            self.display_method(
                note_count=self.frets,
                start_note=self.tuning,
                yield_all=True,
            )
        ):

            fret_value = ''
            if note:
                note_fg = 'green' if note ** self.key.root else 'magenta'
                fret_value = '{}{}{}'.format(
                    FString(note.chr, size=1, fg=note_fg, fx=['']),
                    FString(note.repr_alt, size=0, fg=note_fg, fx=['']),
                    FString(
                        note.repr_oct if self.verbose > 0 else '',
                        size=1,
                        fg=note_fg,
                        fx=['faint' if self.verbose < 2 else ''],
                    ),
                )

            # APPEND NOTE_INFO
            string_line.append(
                FString(
                    fret_value,
                    size=_NOTE_WIDTH -1 if fret_n == 0 else _NOTE_WIDTH,
                    align='cr',
                )
            )

            # APPEND FRET
            string_line.append(
                FString(                   
                    '║' if fret_n % 12 == 0 or fret_n == self.frets -1 else '|', #¦
                    size=_FRET_WIDTH,
                    # fg='blue',
                    # fx=['faint'],
                )
            )

            if fret_n == self.frets -1:
                break

        return str(string_line)


class PluckedStringInstrument(object):

    def __init__(self, *notes, name=''):
        self.name = name
        self.strings = [ PluckedString(*n) for n in notes ]


    def string(self, s):
        return self.strings[s -1]


    def fretboard(self, display=None, frets=12, verbose=1, limit=24):

        if frets > limit:
            frets = limit

        # INIT A NEW SCALE, OTHERWISE YOU USE THE SAME OUTER OBJECT!!!
        if verbose:
            echo(self.fret_inlays(verbose=verbose, frets=frets))

        echo(self.binding('upper', frets=frets))

        for string in self.strings:

            # string numbers
            string_n = FString(
                self.strings.index(string) +1,
                fg='cyan', 
                fx=['faint' if verbose < 1 else ''],
            )

            string.display_method = display
            string.frets = frets
            string.verbose = verbose

            echo(str(string_n) + str(string))
        
        echo(self.binding('lower', frets=frets))


    @classmethod
    def fret_inlays(cls, verbose=1, frets=12):

        if not verbose:
            return ''

        inlays = (
            '',
            'I' if verbose > 1 else '',
            'II' if verbose > 1 else '',
            'III',
            'IV' if verbose > 1 else '',
            'V',
            'VI' if verbose > 1 else '',
            'VII',
            'VIII' if verbose > 1 else '',
            'IX',
            'X' if verbose > 1 else '',
            'XI' if verbose > 1 else '',

            'XII',
            'XIII' if verbose > 1 else '',
            'XIV' if verbose > 1 else '',
            'XV',
            'XVI' if verbose > 1 else '',
            'XVII',
            'XVIII' if verbose > 1 else '',
            'XIX',
            'XX' if verbose > 1 else '',
            'XXI',
            'XXII' if verbose > 1 else '',
            'XXIII' if verbose > 1 else '',

            'XXIV'
        )

        inlay_row = Row(
            FString(
                inlays[0],
                size=_NOTE_WIDTH + _FRET_WIDTH,
                align='l'
            ),
            size=6 +83 # why this??
        )

        i = 1
        while frets:
            inlay_row.append(
                FString(
                    inlays[i],
                    size=_NOTE_WIDTH + _FRET_WIDTH,
                    align='cl',
                    fg='cyan',
                    fx=['faint' if i in (1, 2, 4, 6, 8, 10, 11, 13, 14, 16, 18, 20, 22, 23) else ''],
                )
            )
            frets -= 1
            i += 1

        return inlay_row


    @staticmethod
    def binding(side, frets=12):
        binding = {'upper': '═', 'lower': '═'}
        capo = {'upper': '     ╔', 'lower': '     ╚'}
        fine = {'upper': '╗', 'lower': '╝'}
        fret_binding = binding[side] * frets * (_NOTE_WIDTH + _FRET_WIDTH)
        return capo[side] + fret_binding[:-1] + fine[side]




from bestia.output import echo

class Degrees(object):

    def __init__(self, *degrees):
        self.original_length = len(degrees)
        self.original_order  = tuple(degrees)
        self.reset()

    def __repr__(self):
        return str(self.current_order)

    def __getitem__(self, i):
        return self.current_order[i]


    def reset(self):
        ''' ALWAYS iterate thru items in original_order
            BUT do not modify its contents
        '''
        self.current_order = list(self.original_order)

    def rotate(self):
        self.current_order.insert(
            # pop item0 into last_item
            self.original_length, self.current_order.pop(0)
        )

    def rotate_by_note(self, note):
        if note in self.original_order:
            return self._rotate_by_presence(note)
        return self._rotate_by_char(note)
        # return self._rotate_by_magnitude(note)

    def _rotate_by_presence(self, note):
        ''' exact Note in order:
            == enforces exact note match
        '''
        self.reset()
        for degree in self.original_order:
            if degree == note:
                return True
            self.rotate()

    def _rotate_by_magnitude(self, note):
        ''' exact Note NOT in order:
            >= allows enharmonic equality
            octs from original_order degrees MUST BE
            increased by note.oct for >= evaluation
        '''
        self.reset()
        for degree in self.original_order:

            degree.oct += note.oct
            input(degree)
            if degree >= note:
                degree.oct -= note.oct
                return True

            degree.oct -= note.oct
            self.rotate()

    def _rotate_by_char(self, note):

        self.reset()

        for d, _ in enumerate(self.original_order):

            # B0
            for rc in self.degree_char_check(d, note):

                echo(self.original_order[d], 'red')
                input()

                # try degree's next adj_char
                if rc == 'next_d':
                    self.rotate()
                    break

                # match check note
                elif rc == 'match':
                    return True

                # reached next degree_order.chr
                # elif rc == 'next_c':
                #     pass


    def degree_char_check(self, degree=0, note=None):

        this_deg = self.original_order[degree]
        next_deg = self.original_order[degree +1]

        c = 1
        while True:

            echo(this_deg, 'green')
            input()

            if this_deg.adjacent_chr(c) == note.chr:
                # match check note
                yield 'match'

            if this_deg.adjacent_chr(c) == next_deg.chr:
                # reached next degree_order.chr
                yield 'next_d'

            # try degree's next adj_char
            # yield 'next_c'
            c += 1



from bestia.output import Row, FString, echo

from .degrees import *
from .notes import *

class TonalKey(object):

    def __init__(self, chr, alt='', oct=0):
        self.root = Note(chr, alt, 0) # ignore note.oct

    def __repr__(self):
        spell_line = Row()
        for d in self.scale(
            note_count=len(self.root_intervals) +1, yield_all=False
        ):
            spell_line.append(
                FString(d, size=5)
            )
        return str(spell_line)

    def __getitem__(self, i):
        return self.degree(i)

    @property
    def name(self):
        return self.__class__.__name__

    @classmethod
    def __possible_root_notes(cls, valids=True):
        ''' checks all possible notes for validity as root of given key '''
        valid_roots = []
        invalid_roots = []

        for note in notes_by_alts():

            try:
                invalid_root = False
                for _ in cls(*note).scale(note_count=len(cls.root_intervals) +1, yield_all=0):
                    # if any degree fails, scale is not spellable
                    pass

            except InvalidNote:
                invalid_root = True

            finally:
                if invalid_root:
                    invalid_roots.append(note)
                else:
                    valid_roots.append(note)

        return valid_roots if valids else invalid_roots

    @classmethod
    def valid_root_notes(cls):
        ''' returns only valid root notes for given key class '''
        return cls.__possible_root_notes(valids=True)

    @classmethod
    def invalid_root_notes(cls):
        ''' returns only invalid root notes for given key class '''
        return cls.__possible_root_notes(valids=False)

    def degree_root_interval(self, d):
        ''' return degree's delta semitones from key's root '''
        if d > len(self.root_intervals):
            return self.degree_root_interval(
                d - len(self.root_intervals)
            ) + OCTAVE
        return self.root_intervals[d -1]


    def _spell(self, note_count=-1, start_note=None, yield_all=True, degree_order=[]):
        ''' supported features:
                * positive note_count yields notes exactly
                * negative note_count yields notes indefinetly
        '''

        for note in self._filter_degrees( start_note=start_note, yield_all=yield_all, degree_order=degree_order ):

            if note_count == 0:  # negative note_counts never return
                return

            if yield_all:
                yield note

            elif not yield_all:
                if note:
                    yield note

            if note:
                note_count -= 1


    def _filter_degrees(self, start_note=None, yield_all=True, degree_order=[]):
        ''' supported features:
                * degree order is enforced for chords, modes, etc
        '''

        start_note = start_note if start_note else self.root

        # LOOPED LIST IMPORTED FROM NOTES.PY

        if not degree_order:
            degs = LoopedList(
                * [ self.degree(n +1) for n in range(len(self.root_intervals)) ]
            )
        else:
            degs = LoopedList(
                * [ self.degree(n) for n in degree_order ]
            )

        # rotate degree order by comparing it to star_note
        # degrees = Degrees(*degs)
        # input(degrees.original_order)
        # input(degrees.current_order)
        # degrees.rotate_by_note(start_note)
        # input(degrees.original_order)
        # input(degrees.current_order)

        for note in self._solmizate(
            start_note=start_note, yield_all=yield_all,
        ):

            # THIS.............
            note_match = False

            if note:
                for deg in degs:
                    if note ** deg:
                        yield note
                        note_match = True
                        break

            if not note_match:
                if yield_all:
                    yield None

            # OR THIS.............
            # if note and note ** degrees[0]: # same note, ignr oct
            #     yield note
            #     degrees.rotate()
            #     continue

            # if yield_all:
            #     yield None



    def _solmizate(self, start_note=None, yield_all=True):
        ''' supported features:
                * yields indefinetly
                * yields None for non-diatonic semitones
                * changes octs { degree() }
                * diatonic start_note
                * non-diatonic start_note
        '''
        start_note = start_note if start_note else self.root

        d = 0
        while True:
            d += 1

            # START_NOTE REACHED
            if self[d] < start_note:
                continue

            if yield_all:

                # DONT YIELD EXTRA NONE BEFORE START_NOTE
                # WHEN SCALE DEG BEFORE IS > 1 ST AWAY
                if self[d] != start_note:

                    # YIELD NON-DIATONIC STs, EXCEPT FOR ROOT
                    last_interval = 0 if d == 1 else self[d] - self[d -1] - 1
                    for st in range(last_interval):
                        yield

            yield self[d]


    def scale(self, note_count=0, start_note=None, yield_all=True):
        return self._spell(
            note_count=note_count, start_note=start_note,
            yield_all=yield_all, degree_order=range(1, len(self.root_intervals) +1),
        )


class DiatonicKey(TonalKey):

    def degree(self, d):

        if d < 1:
            return

        if d == 1:
            return self.root

        # GET DEGREE's ROOT OFFSETS = OCTAVE + SPARE_STS
        octs_from_root, spare_sts = divmod(
            self.degree_root_interval(d), OCTAVE
        )
        deg_oct = octs_from_root

        # GET DEGREE PROPERTIES FROM ENHARMONIC MATRIX
        next_degrees = [
            n for n in EnharmonicMatrix[
                self.root.enharmonic_row + spare_sts
            ] if n.chr == self.root.adjacent_chr(d -1) # EXPECTED TONE
        ]

        if len(next_degrees) == 1:
            deg = next_degrees[0]

            # AT THIS POINT DEG_OCT CAN EITHER STAY | +1
            if self.root.oversteps_oct(deg):
                deg_oct += 1

            # RETURN NEW OBJECT, DO NOT CHANGE ENHARMONIC MATRIX ITEM!
            return Note(deg.chr, deg.alt, deg_oct)

        raise InvalidNote


    def triad(self, note_count=0, start_note=None, yield_all=True):
        return self._spell(
            note_count=note_count, start_note=start_note,
            yield_all=yield_all, degree_order=(1, 3, 5),
        )

    def seventh(self, note_count=0, start_note=None, yield_all=True):
        return self._spell(
            note_count=note_count, start_note=start_note,
            yield_all=yield_all, degree_order=(1, 3, 5, 7),
        )

    def ninth(self, note_count=0, start_note=None, yield_all=True):
        return self._spell(
            note_count=note_count, start_note=start_note,
            yield_all=yield_all, degree_order=(1, 3, 5, 7, 9),
        )

    def eleventh(self, note_count=0, start_note=None, yield_all=True):
        return self._spell(
            note_count=note_count, start_note=start_note,
            yield_all=yield_all, degree_order=(1, 3, 5, 7, 9, 11),
        )

    def thirteenth(self, note_count=0, start_note=None, yield_all=True):
        return self._spell(
            note_count=note_count, start_note=start_note,
            yield_all=yield_all, degree_order=(1, 3, 5, 7, 9, 11, 13),
        )


########################
### MAJOR KEYS/MODES ###
########################

class MajorKey(DiatonicKey):

    root_intervals = (
        UNISON,
        MAJOR_SECOND,
        MAJOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MAJOR_SIXTH,
        MAJOR_SEVENTH,
    )

class IonianMode(MajorKey):
    pass

class MixolydianMode(MajorKey):

    root_intervals = (
        UNISON,
        MAJOR_SECOND,
        MAJOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MAJOR_SIXTH,
        MINOR_SEVENTH, # <<<
    )

class LydianMode(MajorKey):

    root_intervals = (
        UNISON,
        MAJOR_SECOND,
        MAJOR_THIRD,
        AUGMENTED_FOURTH, # <<<
        PERFECT_FIFTH,
        MAJOR_SIXTH,
        MAJOR_SEVENTH,
    )


########################
### MINOR KEYS/MODES ###
########################

class MinorKey(DiatonicKey):

    root_intervals = (
        UNISON,
        MAJOR_SECOND,
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MINOR_SIXTH,
        MINOR_SEVENTH,
    )
    
class MinorPentatonicKey(MinorKey):

    root_intervals = (
        UNISON,
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MINOR_SEVENTH,
    )

class Hokkaido(MinorKey):

    root_intervals = (
        UNISON,
        MAJOR_SECOND,
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MINOR_SIXTH,
    )


class NaturalMinorKey(MinorKey):
    pass

class MelodicMinorKey(MinorKey):

    root_intervals = (
        UNISON,
        MAJOR_SECOND,
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MAJOR_SIXTH, # <<<
        MAJOR_SEVENTH, # <<<
    )

class HarmonicMinorKey(MinorKey):

    root_intervals = (
        UNISON,
        MAJOR_SECOND,
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MINOR_SIXTH,
        MAJOR_SEVENTH, # <<<
    )

class AeolianMode(MinorKey):
    pass

class DorianMode(MinorKey):

    root_intervals = (
        UNISON,
        MAJOR_SECOND,
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MAJOR_SIXTH, # <<<
        MINOR_SEVENTH,
    )

class PhrygianMode(MinorKey):

    root_intervals = (
        UNISON,
        MINOR_SECOND, # <<<
        MINOR_THIRD,
        PERFECT_FOURTH,
        PERFECT_FIFTH,
        MINOR_SIXTH,
        MINOR_SEVENTH,
    )


#####################
### CHROMATIC KEY ###
#####################

class ChromaticKey(TonalKey):

    root_intervals = (
        UNISON,
        MINOR_SECOND,
        MAJOR_SECOND,
        MINOR_THIRD,
        MAJOR_THIRD,
        PERFECT_FOURTH,
        AUGMENTED_FOURTH,
        PERFECT_FIFTH,
        MINOR_SIXTH,
        MAJOR_SIXTH,
        MINOR_SEVENTH,
        MAJOR_SEVENTH,
    )

    def degree(self, d):

        if d < 1:
            return

        if d == 1:
            return self.root

        # GET DEGREE's ROOT OFFSETS = OCTAVE + SPARE_STS
        octs_from_root, spare_sts = divmod(
            self.degree_root_interval(d), OCTAVE
        )
        deg_oct = octs_from_root

        # GET DEGREE PROPERTIES FROM ENHARMONIC MATRIX
        # DO I REALLY NEED THESE 3 CHECKS ?
        # MATCH ROOT_TONE
        next_degrees = [
            n for n in EnharmonicMatrix[
                self.root.enharmonic_row + spare_sts
            ] if n ** self.root
        ]

        if not next_degrees:
            # MATCH ROOT_ALT

            next_degrees = [
                n for n in EnharmonicMatrix[
                    self.root.enharmonic_row + spare_sts
                ] if n.alt == self.root.alt[:-1]
            ]

            if not next_degrees:
                # CHOOSE "#" or ""
                chosen_alt = '#' if self.root.alt == '' else self.root.alt
                next_degrees = [
                    n for n in EnharmonicMatrix[
                        self.root.enharmonic_row + spare_sts
                    ] if n.alt == chosen_alt
                ]

        if len(next_degrees) == 1:

            deg = next_degrees[0] # got from ENH_MATRIX

            # AT THIS POINT DEG_OCT CAN EITHER STAY | +1
            if self.root.oversteps_oct(deg):
                deg_oct += 1

            # RETURN NEW OBJECT, DO NOT CHANGE ENHARMONIC MATRIX ITEM!
            return Note(deg.chr, deg.alt, deg_oct)

        raise InvalidNote

TONAL_CLASSES = {

    'major': MajorKey,

    'minor': MinorKey,
    'natural_minor': NaturalMinorKey,
    'melodic_minor': MelodicMinorKey,
    'harmonic_minor': HarmonicMinorKey,

    'ionian': IonianMode,
    'lydian': LydianMode,
    'mixo': MixolydianMode,
    'aeolian': AeolianMode,
    'dorian': DorianMode,
    'phrygian': PhrygianMode,

    # 'hokkaido': Hokkaido,

    'chromatic': ChromaticKey,

}

# CHORDS = {
#     # TRIADS ########################
#     '': MajorTriadChord,
#     'M': MajorTriadChord,

#     'm': MinorTriadChord,

#     'a': AugmentedTriadChord,
    #  '+': AugmentedTriadChord,

#     'd': DiminishedTriadChord,
#     '-': DiminishedTriadChord,

#     # SEVENTH #######################
#     '7': DominantSeventhChord,
#     'dom7': DominantSeventhChord,

#     'M7': MajorSeventhChord,
#     'maj7': MajorSeventhChord,
#     '△7': MajorSeventhChord,

#     'm7': MinorSeventhChord,
#     'min7': MinorSeventhChord,
    
#     'D7': DiminishedSeventhChord,
#     'dim7': DiminishedSeventhChord,
#     '°7': DiminishedSeventhChord,

#     'd7': HalfDiminishedSeventhChord, # isnt this one usually used for diminished (instead of D7)?
#     'm7-5': HalfDiminishedSeventhChord,
#     '⦰7': HalfDiminishedSeventhChord,

# }



import unittest
import time
import random

from bestia.output import echo

from .instruments import *

def dbg(t='Waiting...', c='magenta'):
    echo(f'** {t} **', c)
    input()

class KeyValidityTest(unittest.TestCase):

    ''' for a given key, allows to verify:
            * invalid roots        
            * octave changes
    '''

    def setUp(self):
        print()

        self.chromatic_key = ChromaticKey
        # test no invalid roots

        self.major_key = MajorKey
        # F𝄫¹ invalid MajorKey
        # B𝄪¹ invalid MajorKey
        # D𝄪¹ invalid MajorKey
        # E𝄪¹ invalid MajorKey
        # G𝄪¹ invalid MajorKey
        # A𝄪¹ invalid MajorKey

        self.minor_key = NaturalMinorKey
        # D𝄫² invalid NaturalMinorKey
        # F𝄫¹ invalid NaturalMinorKey
        # G𝄫¹ invalid NaturalMinorKey
        # C𝄫² invalid NaturalMinorKey
        # B𝄪¹ invalid NaturalMinorKey
        # E𝄪¹ invalid NaturalMinorKey

        self.mel_minor_key = MelodicMinorKey
        # F𝄫¹ invalid MelodicMinorKey
        # G𝄫¹ invalid MelodicMinorKey
        # C𝄫² invalid MelodicMinorKey
        # B𝄪¹ invalid MelodicMinorKey
        # D𝄪¹ invalid MelodicMinorKey
        # E𝄪¹ invalid MelodicMinorKey
        # G𝄪¹ invalid MelodicMinorKey
        # A𝄪¹ invalid MelodicMinorKey

        self.har_minor_key = HarmonicMinorKey
        # D𝄫² invalid HarmonicMinorKey
        # F𝄫¹ invalid HarmonicMinorKey
        # G𝄫¹ invalid HarmonicMinorKey
        # C𝄫² invalid HarmonicMinorKey
        # B𝄪¹ invalid HarmonicMinorKey
        # D𝄪¹ invalid HarmonicMinorKey
        # E𝄪¹ invalid HarmonicMinorKey
        # G𝄪¹ invalid HarmonicMinorKey
        # A𝄪¹ invalid HarmonicMinorKey


    def testValidRoots(self):

        for Key in [self.major_key,self.minor_key ]:

            echo('\nValid {}s'.format(Key.__name__), 'underline')

            for note in Key.valid_root_notes():

                line = Row()
                key = Key(*note)
                for d in key.scale(
                    note_count=len(key.root_intervals) +16, yield_all=False
                ):
                    line.append(
                        FString(
                            d,
                            size=5,
                            fg='blue' if not (d.oct % 2) else 'white', 
                        )
                    )
                line.echo()



    def testInvalidRoots(self):
        echo('\nInvalid {}s'.format(self.major_key.__name__), 'underline')
        for note in self.major_key.invalid_root_notes():
            echo(
                '{}{} invalid {}'.format(note.chr, note.repr_alt, self.major_key.__name__),
                'red', 'faint'
            )



class NoteEqualityTest(unittest.TestCase):

    DANGEROUS_NON_EQUALS = (
        # ''' Used mainly to test B#, Cd, etc... '''

        (Note('C', 'b', 3), Note('B', '#', 3)),

        (Note('C', 'b', 3), Note('B', '#', 3)),

        (Note('E', 'b', 5), Note('D', '#', 4)),

        (Note('B', '', 5), Note('C', 'b', 4)),

        (Note('E', '#', 3), Note('F', '', 4)),


        (Note('B', 'b'), Note('C', 'bb')),
        (Note('A', '#'), Note('C', 'bb')),

        (Note('B'     ), Note('C', 'b')),
        (Note('A', '##'), Note('C', 'b')),

        (Note('B', '#'), Note('C', '')),
        (Note('B', '#'), Note('D', 'bb')),

        (Note('B', '##'), Note('C', '#')),
        (Note('B', '##'), Note('D', 'b')),


        # these should eval False OK
        (Note('A', '#'), Note('B', 'b', 4)),
        (Note('A', '##'), Note('B', '', 4)),
        (Note('C', ''), Note('D', 'bb', 4)),
        (Note('C', '#'), Note('D', 'b', 4)),
    )

    def setUp(self):
        print()

    def testNotEnharmonic(self):
        for note_pair in self.DANGEROUS_NON_EQUALS:
            assert not note_pair[0] << note_pair[1], (note_pair[0], note_pair[1])
            assert not note_pair[0] >> note_pair[1], (note_pair[0], note_pair[1])
            assert note_pair[0] - note_pair[1] != 0, (note_pair[0], note_pair[1])


    def testAreEnharmonic(self):
        ''' checks note_pairs in enharmonic rows
            for different types of enharmonic equality '''
        for note_pair in EnharmonicMatrix:

            # DO NOT PASS STRICT EQUALITY
            assert not note_pair[0] == note_pair[1], (note_pair[0], note_pair[1])

            # PASS ENHARMONIC EQUALITY
            assert note_pair[0] << note_pair[1], (note_pair[0], note_pair[1])
            assert note_pair[0] >> note_pair[1], (note_pair[0], note_pair[1])

            # CHECK DELTA_ST === 0
            assert not note_pair[0] < note_pair[1], (note_pair[0], note_pair[1])
            assert not note_pair[0] > note_pair[1], (note_pair[0], note_pair[1])

            assert note_pair[0] <= note_pair[1], (note_pair[0], note_pair[1])
            assert note_pair[0] >= note_pair[1], (note_pair[0], note_pair[1])

            assert note_pair[0] - note_pair[1] == 0, (note_pair[0], note_pair[1])


    def testEnharmonicOperators(self):
        ''' '''
        print('Testing enharmonic operators')
        Cs5 = Note('C', '#', 5)
        Db3 = Note('D', 'b', 3)
        Cs3 = Note('C', '#', 3)
        assert Cs3 ** Cs5      # same pitch, OLD is_a(oct=None)
        assert Db3 << Cs3      # enh note
        assert Db3 >> Cs3      # enh note
        assert Cs3 << Db3      # enh note
        assert Cs3 >> Db3      # enh note
        assert not Db3 == Cs3  # same note


        # // enhr note, ignr oct
        # ** same note, ignr oct
        # == same note, same oct
        # << enhr note, same oct


class ChromaticKeysTest(unittest.TestCase):

    def setUp(self):
        print()
        self.c_chromatic = ChromaticKey('C')
        self.f_sharp_chromatic = ChromaticKey('F', '#')
        self.b_flat_chromatic = ChromaticKey('B', 'b')

    def testIntervalsCount(self):
        assert len(self.c_chromatic.root_intervals) == 12, self.c_chromatic.root_intervals
        assert len(self.f_sharp_chromatic.root_intervals) == 12, self.f_sharp_chromatic.root_intervals
        assert len(self.b_flat_chromatic.root_intervals) == 12, self.b_flat_chromatic.root_intervals


    def testCChromaticScaleGenerator(self):

        octaves_to_test = 18
        intervals = 12
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 12 + 1 = 217

        for i, note in enumerate(
            self.c_chromatic.scale(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note == Note('C', '', 0), note
                # assert note == Note('C', '', 0), note
            elif i == 2:
                assert note == Note('C', '#', 0), note
            elif i == 3:
                assert note == Note('D', '', 0), note
            elif i == 4:
                assert note == Note('D', '#', 0), note
            elif i == 5:
                assert note == Note('E', '', 0), note
            elif i == 6:
                assert note == Note('F', '', 0), note
            elif i == 7:
                assert note == Note('F', '#', 0), note
            elif i == 8:
                assert note == Note('G', '', 0), note
            elif i == 9:
                assert note == Note('G', '#', 0), note
            elif i == 10:
                assert note == Note('A', '', 0), note
            elif i == 11:
                assert note == Note('A', '#', 0), note
            elif i == 12:
                assert note == Note('B', '', 0), note
            elif i == 13:
                assert note == Note('C', '', 1), note
            elif i == 14:
                assert note == Note('C', '#', 1), note
            elif i == 15:
                assert note == Note('D', '', 1), note
            elif i == 16:
                assert note == Note('D', '#', 1), note
            elif i == 17:
                assert note == Note('E', '', 1), note
            elif i == 18:
                assert note == Note('F', '', 1), note
            elif i == 19:
                assert note == Note('F', '#', 1), note
            elif i == 20:
                assert note == Note('G', '', 1), note
            elif i == 21:
                assert note == Note('G', '#', 1), note
            elif i == 22:
                assert note == Note('A', '', 1), note
            elif i == 23:
                assert note == Note('A', '#', 1), note
            elif i == 24:
                assert note == Note('B', '', 1), note
            elif i == 25:
                assert note == Note('C', '', 2), note
            # ..............................
            elif i == 97:
                assert note == Note('C', '', 8), note
            elif i == 98:
                assert note == Note('C', '#', 8), note
            elif i == 99:
                assert note == Note('D', '', 8), note
            elif i == 100:
                assert note == Note('D', '#', 8), note
            elif i == 101:
                assert note == Note('E', '', 8), note
            elif i == 102:
                assert note == Note('F', '', 8), note
            elif i == 103:
                assert note == Note('F', '#', 8), note
            elif i == 104:
                assert note == Note('G', '', 8), note
            elif i == 105:
                assert note == Note('G', '#', 8), note
            elif i == 106:
                assert note == Note('A', '', 8), note
            elif i == 107:
                assert note == Note('A', '#', 8), note
            elif i == 108:
                assert note == Note('B', '', 8), note
            elif i == 109:
                assert note == Note('C', '', 9), note
            # ..............................
            elif i == 205:
                assert note == Note('C', '', 17), note
            elif i == 206:
                assert note == Note('C', '#', 17), note
            elif i == 207:
                assert note == Note('D', '', 17), note
            elif i == 208:
                assert note == Note('D', '#', 17), note
            elif i == 209:
                assert note == Note('E', '', 17), note
            elif i == 210:
                assert note == Note('F', '', 17), note
            elif i == 211:
                assert note == Note('F', '#', 17), note
            elif i == 212:
                assert note == Note('G', '', 17), note
            elif i == 213:
                assert note == Note('G', '#', 17), note
            elif i == 214:
                assert note == Note('A', '', 17), note
            elif i == 215:
                assert note == Note('A', '#', 17), note
            elif i == 216:
                assert note == Note('B', '', 17), note
            elif i == 217:
                assert note == Note('C', '', 18), note


    def testFSharpChromaticScaleGenerator(self):

        octaves_to_test = 18
        intervals = 12
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 12 + 1 = 217

        for i, note in enumerate(
            self.f_sharp_chromatic.scale(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note == Note('F', '#', 0), note
            elif i == 2:
                assert note == Note('G', '', 0), note
            elif i == 3:
                assert note == Note('G', '#', 0), note
            elif i == 4:
                assert note == Note('A', '', 0), note
            elif i == 5:
                assert note == Note('A', '#', 0), note
            elif i == 6:
                assert note == Note('B', '', 0), note
            elif i == 7:
                assert note == Note('C', '', 1), note
            elif i == 8:
                assert note == Note('C', '#', 1), note
            elif i == 9:
                assert note == Note('D', '', 1), note
            elif i == 10:
                assert note == Note('D', '#', 1), note
            elif i == 11:
                assert note == Note('E', '', 1), note
            elif i == 12:
                assert note == Note('F', '', 1), note
            elif i == 13:
                assert note == Note('F', '#', 1), note
            elif i == 14:
                assert note == Note('G', '', 1), note
            elif i == 15:
                assert note == Note('G', '#', 1), note
            elif i == 16:
                assert note == Note('A', '', 1), note
            elif i == 17:
                assert note == Note('A', '#', 1), note
            elif i == 18:
                assert note == Note('B', '', 1), note
            elif i == 19:
                assert note == Note('C', '', 2), note
            # ..............................
            elif i == 211:
                assert note == Note('C', '', 18), note
            elif i == 212:
                assert note == Note('C', '#', 18), note
            elif i == 213:
                assert note == Note('D', '', 18), note
            elif i == 214:
                assert note == Note('D', '#', 18), note
            elif i == 215:
                assert note == Note('E', '', 18), note
            elif i == 216:
                assert note == Note('F', '', 18), note
            elif i == 217:
                assert note == Note('F', '#', 18), note



    def testBFlatChromaticScaleGenerator(self):

        octaves_to_test = 18
        intervals = 12
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 12 + 1 = 217

        for i, note in enumerate(
            self.b_flat_chromatic.scale(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note == Note('B', 'b', 0), note
            elif i == 2:
                assert note == Note('B', '', 0), note
            elif i == 3:
                assert note == Note('C', '', 1), note
            elif i == 4:
                assert note == Note('D', 'b', 1), note
            elif i == 5:
                assert note == Note('D', '', 1), note
            elif i == 6:
                assert note == Note('E', 'b', 1), note
            elif i == 7:
                assert note == Note('E', '', 1), note
            elif i == 8:
                assert note == Note('F', '', 1), note
            elif i == 9:
                assert note == Note('G', 'b', 1), note
            elif i == 10:
                assert note == Note('G', '', 1), note
            elif i == 11:
                assert note == Note('A', 'b', 1), note
            elif i == 12:
                assert note == Note('A', '', 1), note
            elif i == 13:
                assert note == Note('B', 'b', 1), note
            elif i == 14:
                assert note == Note('B', '', 1), note
            elif i == 15:
                assert note == Note('C', '', 2), note
            # ..............................
            elif i == 200:
                assert note == Note('F', '', 17), note
            elif i == 201:
                assert note == Note('G', 'b', 17), note
            elif i == 202:
                assert note == Note('G', '', 17), note
            elif i == 203:
                assert note == Note('A', 'b', 17), note
            elif i == 204:
                assert note == Note('A', '', 17), note
            elif i == 205:
                assert note == Note('B', 'b', 17), note
            elif i == 206:
                assert note == Note('B', '', 17), note
            elif i == 207:
                assert note == Note('C', '', 18), note
            elif i == 208:
                assert note == Note('D', 'b', 18), note
            elif i == 209:
                assert note == Note('D', '', 18), note
            elif i == 210:
                assert note == Note('E', 'b', 18), note
            elif i == 211:
                assert note == Note('E', '', 18), note
            elif i == 212:
                assert note == Note('F', '', 18), note
            elif i == 213:
                assert note == Note('G', 'b', 18), note
            elif i == 214:
                assert note == Note('G', '', 18), note
            elif i == 215:
                assert note == Note('A', 'b', 18), note
            elif i == 216:
                assert note == Note('A', '', 18), note
            elif i == 217:
                assert note == Note('B', 'b', 18), note





class MajorKeysExpectedNotesTest(unittest.TestCase):

    def setUp(self):
        print()
        self.c_major = MajorKey('C')
        self.b_major = MajorKey('B')            # 5 sharps
        self.d_flat_major = MajorKey('D', 'b')  # 5 flats

    def testIntervalsCount(self):
        assert len(self.c_major.root_intervals) == 7, self.c_major.root_intervals
        assert len(self.b_major.root_intervals) == 7, self.b_major.root_intervals
        assert len(self.d_flat_major.root_intervals) == 7, self.d_flat_major.root_intervals


    def testCMajorScaleGenerator(self):

        octaves_to_test = 18
        intervals = 7
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 7 + 1 = 127

        for i, note in enumerate(
            self.c_major.scale(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note == Note('C', '', 0), note
            elif i == 2:
                assert note == Note('D', '', 0), note
            elif i == 3:
                assert note == Note('E', '', 0), note
            elif i == 4:
                assert note == Note('F', '', 0), note
            elif i == 5:
                assert note == Note('G', '', 0), note
            elif i == 6:
                assert note == Note('A', '', 0), note
            elif i == 7:
                assert note == Note('B', '', 0), note
            elif i == 8:
                assert note == Note('C', '', 1), note
            elif i == 9:
                assert note == Note('D', '', 1), note
            elif i == 10:
                assert note == Note('E', '', 1), note
            elif i == 11:
                assert note == Note('F', '', 1), note
            elif i == 12:
                assert note == Note('G', '', 1), note
            elif i == 13:
                assert note == Note('A', '', 1), note
            elif i == 14:
                assert note == Note('B', '', 1), note
            elif i == 15:
                assert note == Note('C', '', 2), note
            # ..............................
            elif i == 64:
                assert note == Note('C', '', 9), note
            elif i == 65:
                assert note == Note('D', '', 9), note
            elif i == 66:
                assert note == Note('E', '', 9), note
            elif i == 67:
                assert note == Note('F', '', 9), note
            elif i == 68:
                assert note == Note('G', '', 9), note
            elif i == 69:
                assert note == Note('A', '', 9), note
            elif i == 70:
                assert note == Note('B', '', 9), note
            elif i == 71:
                assert note == Note('C', '', 10), note
            # ..............................
            elif i == 120:
                assert note == Note('C', '', 17), note
            elif i == 121:
                assert note == Note('D', '', 17), note
            elif i == 122:
                assert note == Note('E', '', 17), note
            elif i == 123:
                assert note == Note('F', '', 17), note
            elif i == 124:
                assert note == Note('G', '', 17), note
            elif i == 125:
                assert note == Note('A', '', 17), note
            elif i == 126:
                assert note == Note('B', '', 17), note
            elif i == 127:
                assert note == Note('C', '', 18), note


    def testBMajorScaleGenerator(self):

        octaves_to_test = 18
        intervals = 7
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 7 + 1 = 127

        for i, note in enumerate(
            self.b_major.scale(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note == Note('B', '', 0), note
            elif i == 2:
                assert note == Note('C', '#', 1), note
            elif i == 3:
                assert note == Note('D', '#', 1), note
            elif i == 4:
                assert note == Note('E', '', 1), note
            elif i == 5:
                assert note == Note('F', '#', 1), note
            elif i == 6:
                assert note == Note('G', '#', 1), note
            elif i == 7:
                assert note == Note('A', '#', 1), note
            elif i == 8:
                assert note == Note('B', '', 1), note
            elif i == 9:
                assert note == Note('C', '#', 2), note
            elif i == 10:
                assert note == Note('D', '#', 2), note
            elif i == 11:
                assert note == Note('E', '', 2), note
            elif i == 12:
                assert note == Note('F', '#', 2), note
            elif i == 13:
                assert note == Note('G', '#', 2), note
            elif i == 14:
                assert note == Note('A', '#', 2), note
            elif i == 15:
                assert note == Note('B', '', 2), note
            elif i == 16:
                assert note == Note('C', '#', 3), note
            # ..............................
            elif i == 64:
                assert note == Note('B', '', 9), note
            elif i == 65:
                assert note == Note('C', '#', 10), note
            elif i == 66:
                assert note == Note('D', '#', 10), note
            elif i == 67:
                assert note == Note('E', '', 10), note
            elif i == 68:
                assert note == Note('F', '#', 10), note
            elif i == 69:
                assert note == Note('G', '#', 10), note
            elif i == 70:
                assert note == Note('A', '#', 10), note
            elif i == 71:
                assert note == Note('B', '', 10), note
            # ..............................
            elif i == 120:
                assert note == Note('B', '', 17), note
            elif i == 121:
                assert note == Note('C', '#', 18), note
            elif i == 122:
                assert note == Note('D', '#', 18), note
            elif i == 123:
                assert note == Note('E', '', 18), note
            elif i == 124:
                assert note == Note('F', '#', 18), note
            elif i == 125:
                assert note == Note('G', '#', 18), note
            elif i == 126:
                assert note == Note('A', '#', 18), note
            elif i == 127:
                assert note == Note('B', '', 18), note



    def testDFlatMajorScaleGenerator(self):

        octaves_to_test = 18
        intervals = 7
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 7 + 1 = 127

        for i, note in enumerate(
            self.d_flat_major.scale(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note == Note('D', 'b', 0), note
            elif i == 2:
                assert note == Note('E', 'b', 0), note
            elif i == 3:
                assert note == Note('F', '', 0), note
            elif i == 4:
                assert note == Note('G', 'b', 0), note
            elif i == 5:
                assert note == Note('A', 'b', 0), note
            elif i == 6:
                assert note == Note('B', 'b', 0), note
            elif i == 7:
                assert note == Note('C', '', 1), note
            elif i == 8:
                assert note == Note('D', 'b', 1), note
            elif i == 9:
                assert note == Note('E', 'b', 1), note
            elif i == 10:
                assert note == Note('F', '', 1), note
            elif i == 11:
                assert note == Note('G', 'b', 1), note
            elif i == 12:
                assert note == Note('A', 'b', 1), note
            elif i == 13:
                assert note == Note('B', 'b', 1), note
            elif i == 14:
                assert note == Note('C', '', 2), note
            elif i == 15:
                assert note == Note('D', 'b', 2), note
            elif i == 16:
                assert note == Note('E', 'b', 2), note
            # ..............................
            elif i == 64:
                assert note == Note('D', 'b', 9), note
            elif i == 65:
                assert note == Note('E', 'b', 9), note
            elif i == 66:
                assert note == Note('F', '', 9), note
            elif i == 67:
                assert note == Note('G', 'b', 9), note
            elif i == 68:
                assert note == Note('A', 'b', 9), note
            elif i == 69:
                assert note == Note('B', 'b', 9), note
            elif i == 70:
                assert note == Note('C', '', 10), note
            elif i == 71:
                assert note == Note('D', 'b', 10), note
            elif i == 72:
                assert note == Note('E', 'b', 10), note
            # ..............................
            elif i == 120:
                assert note == Note('D', 'b', 17), note
            elif i == 121:
                assert note == Note('E', 'b', 17), note
            elif i == 122:
                assert note == Note('F', '', 17), note
            elif i == 123:
                assert note == Note('G', 'b', 17), note
            elif i == 124:
                assert note == Note('A', 'b', 17), note
            elif i == 125:
                assert note == Note('B', 'b', 17), note
            elif i == 126:
                assert note == Note('C', '', 18), note
            elif i == 127:
                assert note == Note('D', 'b', 18), note


    def testDegreeMethod(self):
        ''' tests __getitem__ == degree() '''
        degree = self.c_major.degree(1)
        item = self.c_major[1]
        note = Note('C', 0)
        assert degree == item, (degree, item)
        assert degree == note, (degree, note)



class TonalKeySpellMethodTest(unittest.TestCase):

    def setUp(self):
        print()
        self.keys = {
            'Ab_chromatic_key': ChromaticKey('A', 'b'),
            'B_major': MajorKey('B'),
            'Bb_minor': NaturalMinorKey('B', 'b'),
            'C_mel_minor': MelodicMinorKey('C'),
            'F#_har_minor': HarmonicMinorKey('F', '#'),
            # 'E7': SeventhDominant('E'),   # test chords too eventually
        }

    def testNoteCount(self):
        ''' tests yielded note count is what has been required '''
        for key in self.keys.values():
            max_notes = random.randint(2, 64)
            print(f'Testing {key.root.chr}{key.root.repr_alt} {key.__class__.__name__}._spell( note_count=1..{max_notes} ) ...')
            for count in range(max_notes):
                count += 1
                yielded_notes = len(
                    [ n for n in key._spell(
                        note_count=count, start_note=None, yield_all=False, degree_order=[]
                    ) ]
                )
                assert yielded_notes == count, (yielded_notes, count)


    def testDiatonicStartNote(self):
        ''' tests that first yielded note == diatonic start_note '''
        for key in self.keys.values():
            d = random.randint(2, 64)
            print(f'Testing {key.root.chr}{key.root.repr_alt} {key.__class__.__name__}._spell( start_note = degree({d}) ) ...')
            for note in key._spell(
                note_count=1, start_note=key.degree(d), yield_all=True, degree_order=[]
            ): 
                assert note == Note(*key.degree(d)), note


    def testNonDiatonicStartNoteYieldNotes(self):
        ''' tests 1st yielded item == expected diatonic note when:
                * start_note is non-diatonic to the scale
                * Nones ARE NOT being yielded
        '''
        test_parameters = [

            {
                'key': self.keys['Ab_chromatic_key'], 
                'non_diatonic_note': Note('A', '#', 1), # enharmonic
                'exp_diatonic_note': Note('B', 'b', 1), # equals
            },

            {
                'key': self.keys['B_major'], 
                'non_diatonic_note': Note('D', 1),      # missing note
                'exp_diatonic_note': Note('D', '#', 1), # next note
            },

            {
                'key': self.keys['Bb_minor'], 
                'non_diatonic_note': Note('D', '#', 1), # enharmonic
                'exp_diatonic_note': Note('E', 'b', 1), # equals
            },
            {
                'key': self.keys['C_mel_minor'], 
                'non_diatonic_note': Note('F', '#', 0), # missing note
                'exp_diatonic_note': Note('G', '',  0), # next note
            },
            {
                'key': self.keys['F#_har_minor'], 
                'non_diatonic_note': Note('C', 'b', 4), # enharmonic
                'exp_diatonic_note': Note('B', '' , 3), # equals
            },

            # {   # test chords too eventually
            #     'key': self.keys['E7'], 
            #     'non_diatonic_note': Note('A', '', 0), # missing note
            #     'exp_diatonic_note': Note('B', '', 0), # next note
            # },
        ]

        for param in test_parameters:
            key = param['key']
            non = param['non_diatonic_note']
            exp = param['exp_diatonic_note']
            print(f'Testing {key.root.chr}{key.root.repr_alt} {key.__class__.__name__}._spell( start_note = non_diatonic_note , yield_all = 0 ) ...')
            for note in key._spell(
                note_count=1, start_note=non, yield_all=False, degree_order=[]
            ): 
                assert note != None, type(note)         # yield all=False ensures no Nones
                assert note == Note(*exp), (note, exp)


    def testNonDiatonicStartNoteYieldAll(self):
        ''' tests 1st yielded item == expected value when:
                * start_note is non-diatonic to the scale
                * Nones ARE being yielded
        '''
        test_parameters = [
            {
                'key': self.keys['Ab_chromatic_key'], 
                'non_diatonic_note': Note('A', '#', 1), # enharmonic
                'exp_diatonic_note': Note('B', 'b', 1), # equals
            },
            {
                'key': self.keys['B_major'], 
                'non_diatonic_note': Note('D', 1),      # missing note
                'exp_diatonic_note': None,              # yields a None, not D#
            },
            {
                'key': self.keys['Bb_minor'], 
                'non_diatonic_note': Note('D', '#', 1), # enharmonic
                'exp_diatonic_note': Note('E', 'b', 1), # equals
            },
            {
                'key': self.keys['C_mel_minor'], 
                'non_diatonic_note': Note('F', '#', 0), # missing note
                'exp_diatonic_note': None,              # yields a None, not G
            },
            {
                'key': self.keys['F#_har_minor'], 
                'non_diatonic_note': Note('C', 'b', 4), # enharmonic
                'exp_diatonic_note': Note('B', '' , 3), # equals
            },
        ]

        for param in test_parameters:
            key = param['key']
            non = param['non_diatonic_note']
            exp = param['exp_diatonic_note']
            
            print(f'Testing {key.root.chr}{key.root.repr_alt} {key.__class__.__name__}._spell( start_note = non_diatonic_note , yield_all = 1 ) ...')
            
            for note in key._spell(
                note_count=1, start_note=non, yield_all=True, degree_order=[]
            ):

                assert note == exp, (note, exp)
                # if note is not None:
                #     assert note == Note(*exp), (note, exp)

                break # test only first yielded value even if not a note
                

    def testMajorNoneYields(self):
        for i, note in enumerate(MajorKey('C')._spell(
            note_count=64, start_note=None, yield_all=True, degree_order=[]
        )):
            # echo(note, 'yellow')
            if i == 0:
                assert note == Note('C', '', 0), note
            elif i == 1:
                assert note is None
            elif i == 2:
                assert note == Note('D', '', 0), note
            elif i == 3:
                assert note is None
            elif i == 4:
                assert note == Note('E', '', 0), note
            elif i == 5:
                assert note == Note('F', '', 0), note
            elif i == 6:
                assert note is None
            elif i == 7:
                assert note == Note('G', '', 0), note
            elif i == 8:
                assert note is None
            elif i == 9:
                assert note == Note('A', '', 0), note
            elif i == 10:
                assert note is None
            elif i == 11:
                assert note == Note('B', '', 0), note
            elif i == 12:
                assert note == Note('C', '', 1), note


    def testChromaticNoneYields(self):
        ''' tests Chromatic scale does NEVER yield None even with yield_all '''
        for i, note in enumerate(self.keys['Ab_chromatic_key']._spell(
            note_count=128, start_note=None, yield_all=True, degree_order=[]
        )):
            assert note != None


    def testDegreeOrder(self):
        sn = Note('C', '', 0)
        for i, note in enumerate(MajorKey('C')._spell(
            note_count=23, start_note=sn, yield_all=False, degree_order=[1, 3, 5, 7]
        )):
            echo(f'{note}  ', 'yellow', mode='raw')
            # time.sleep(0.3)


    def testDegreeOrderOverOct(self):
        for i, note in enumerate(MajorKey('C').ninth(
            note_count=6, start_note=None, yield_all=False
        )):
            i += 1
            if i == 1:
                assert note == Note('C', '', 0), note
            # elif i == 2:
            #     assert note == Note('E', '', 0), note
            # elif i == 3:
            #     assert note == Note('G', '', 0), note
            # elif i == 4:
            #     assert note == Note('B', '', 0), note
            # elif i == 5:
            #     assert note == Note('D', '', 1), note
            # elif i == 6:
            #     assert note == Note('C', '', 2), note


